// 
//  (c) Microsoft Corporation. See LICENSE.TXT file for licensing details
//  
namespace BackendWebRole
{
    using System;
    using System.Diagnostics;
    using System.Threading;

    abstract class AsyncResult : IAsyncResult
    {
        readonly AsyncCallback callback;
        readonly object state;
        readonly object thisLock;
        bool completedSynchronously;
        bool endCalled;
        Exception exception;
        bool isCompleted;
        ManualResetEvent manualResetEvent;

        protected AsyncResult(AsyncCallback callback, object state)
        {
            this.callback = callback;
            this.state = state;
            this.thisLock = new object();
        }

        public bool HasCallback { get { return this.callback != null; } }
        object ThisLock { get { return this.thisLock; } }

        public object AsyncState { get { return this.state; } }

        public WaitHandle AsyncWaitHandle
        {
            get
            {
                if (this.manualResetEvent != null)
                {
                    return this.manualResetEvent;
                }

                lock (this.ThisLock)
                {
                    if (this.manualResetEvent == null)
                    {
                        this.manualResetEvent = new ManualResetEvent(this.isCompleted);
                    }
                }

                return this.manualResetEvent;
            }
        }

        public bool CompletedSynchronously { get { return this.completedSynchronously; } }

        public bool IsCompleted { get { return this.isCompleted; } }

        /// <summary>
        ///   Call this version of complete when your asynchronous operation is complete.  This will update the state
        ///   of the operation and notify the callback.
        /// </summary>
        /// <param name = "completedSynchronously"></param>
        protected void Complete(bool completedSynchronously)
        {
            if (this.isCompleted)
            {
                throw new InvalidOperationException();
            }


            this.completedSynchronously = completedSynchronously;

            if (completedSynchronously)
            {
                // If we completedSynchronously, then there's no chance that the manualResetEvent was created so
                // we don't need to worry about a race
                Debug.Assert(this.manualResetEvent == null, "No ManualResetEvent should be created for a synchronous AsyncResult.");
                this.isCompleted = true;
            }
            else
            {
                lock (this.ThisLock)
                {
                    this.isCompleted = true;
                    if (this.manualResetEvent != null)
                    {
                        this.manualResetEvent.Set();
                    }
                }
            }

            if (this.callback != null)
            {
                try
                {
                    this.callback(this);
                }
                catch (Exception e)
                {
                    throw new InvalidOperationException("AsyncCallbackException", e);
                }
            }
        }

        /// <summary>
        ///   Call this version of complete if you raise an exception during processing.  In addition to notifying
        ///   the callback, it will capture the exception and store it to be thrown during AsyncResult.End.
        /// </summary>
        /// <param name = "exception"></param>
        protected void Complete(bool completedSynchronously, Exception exception)
        {
            this.exception = exception;
            this.Complete(completedSynchronously);
        }

        /// <summary>
        ///   End should be called when the End function for the asynchronous operation is complete.  It
        ///   ensures the asynchronous operation is complete, and does some common validation.
        /// </summary>
        /// <param name = "result"></param>
        protected static TAsyncResult End<TAsyncResult>(IAsyncResult result) where TAsyncResult : AsyncResult
        {
            if (result == null)
            {
                throw new ArgumentNullException("result");
            }

            var asyncResult = result as TAsyncResult;

            if (asyncResult == null)
            {
                throw new ArgumentException("result", "InvalidAsyncResult");
            }

            if (asyncResult.endCalled)
            {
                throw new InvalidOperationException("AsyncObjectAlreadyEnded");
            }

            asyncResult.endCalled = true;

            if (!asyncResult.isCompleted)
            {
                asyncResult.AsyncWaitHandle.WaitOne();
            }

            if (asyncResult.manualResetEvent != null)
            {
                asyncResult.manualResetEvent.Close();
            }

            if (asyncResult.exception != null)
            {
                throw asyncResult.exception;
            }

            return asyncResult;
        }
    }
}