﻿// 
//  (c) Microsoft Corporation. See LICENSE.TXT file for licensing details
//  
namespace BackendWebRole
{
    using System;
    using System.Configuration;
    using System.IO;
    using System.Net;
    using System.Security.Cryptography;
    using System.ServiceModel;
    using System.ServiceModel.Channels;
    using System.ServiceModel.Description;
    using System.ServiceModel.Dispatcher;
    using System.Text;
    using Microsoft.ServiceBus;

    [ServiceContract]
    [ServiceBehavior(AddressFilterMode = AddressFilterMode.Any)]
    [DefaultOperation("HandleRequest")]
    public class ServiceBusPassthrough
    {
        static readonly Uri forwardService;
        static readonly byte[] signingKey;
        static readonly TokenProvider tokenProvider;

        static ServiceBusPassthrough()
        {
            forwardService = ServiceBusEnvironment.CreateServiceUri(Uri.UriSchemeHttps, ConfigurationManager.AppSettings["serviceBusNamespace"], String.Empty);
            tokenProvider = TokenProvider.CreateSharedSecretTokenProvider("owner", ConfigurationManager.AppSettings["managementKey"]);
            signingKey = Convert.FromBase64String(ConfigurationManager.AppSettings["sharedSignature"]);
        }

        [OperationContract(Action = "*", ReplyAction = "*", AsyncPattern = true)]
        public IAsyncResult BeginHandleRequest(Message request, AsyncCallback callback, object state)
        {
            var ctx = OperationContext.Current;
            var path = ctx.Channel.LocalAddress.Uri.AbsolutePath;
            var requestPath = request.Properties.Via.PathAndQuery;
            return new HandleRequestAsyncResult(
                request, signingKey, new Uri(forwardService, requestPath.Substring(path.Length)), tokenProvider, callback, state);
        }

        public Message EndHandleRequest(IAsyncResult result)
        {
            return HandleRequestAsyncResult.End(result);
        }

        class HandleRequestAsyncResult : AsyncResult
        {
            readonly string endpoint;
            readonly HttpWebRequest forwarderRequest;
            readonly Message request;
            Message result;

            public HandleRequestAsyncResult(
                Message request, byte[] signingKey, Uri forwardService, TokenProvider tokenProvider, AsyncCallback callback, object state)
                : base(callback, state)
            {
                this.request = request;

                HttpRequestMessageProperty rm;
                if (request.Properties.ContainsKey(HttpRequestMessageProperty.Name) &&
                    this.ValidateRequest(rm = (HttpRequestMessageProperty) request.Properties[HttpRequestMessageProperty.Name]))
                {
                    this.endpoint = rm.Headers["Endpoint"];
                    if (this.endpoint != null)
                    {
                        rm.Headers.Remove("Endpoint");
                    }
                    this.forwarderRequest = (HttpWebRequest) WebRequest.Create(forwardService);
                    this.forwarderRequest.Method = rm.Method;
                    this.MapRequestHeaders(rm);
                    tokenProvider.BeginGetWebToken(
                                                   new UriBuilder(this.forwarderRequest.RequestUri) {Scheme = Uri.UriSchemeHttp}.Uri.AbsoluteUri,
                                                   "Send",
                                                   false,
                                                   OperationContext.Current.Channel.OperationTimeout,
                                                   this.TokenRequestComplete,
                                                   null);
                }
                else
                {
                    this.CompleteWithError(HttpStatusCode.Forbidden, true);
                }
            }

            void TokenRequestComplete(IAsyncResult ar)
            {
                try
                {
                    var token = tokenProvider.EndGetWebToken(ar);
                    this.forwarderRequest.Headers.Add(HttpRequestHeader.Authorization, token);
                    this.forwarderRequest.BeginGetResponse(this.RequestCompleted, null);
                }
                catch (Exception e)
                {
                    this.Complete(false, e);
                }
            }

            void RequestCompleted(IAsyncResult ar)
            {
                try
                {
                    var response = (HttpWebResponse) this.forwarderRequest.EndGetResponse(ar);
                    var responseProperty = new HttpResponseMessageProperty();
                    responseProperty.Headers.Add(response.Headers);
                    responseProperty.StatusCode = response.StatusCode;
                    responseProperty.StatusDescription = response.StatusDescription;
                    responseProperty.SuppressEntityBody = true;
                    var location = responseProperty.Headers["Location"];
                    if (location != null)
                    {
                        var ep = this.endpoint.Split(':');
                        var uriBuilder = new UriBuilder(location);
                        uriBuilder.Scheme = Uri.UriSchemeHttp;
                        uriBuilder.Path = "/Passthrough.svc" + uriBuilder.Path;
                        uriBuilder.Host = ep[0];
                        if (ep.Length > 1)
                        {
                            uriBuilder.Port = int.Parse(ep[1]);
                        }
                        responseProperty.Headers["Location"] = uriBuilder.Uri.AbsoluteUri;
                    }

                    this.SignResponse(responseProperty);
                    this.result = Message.CreateMessage(this.request.Version, String.Empty);
                    this.result.Properties.Add(HttpResponseMessageProperty.Name, responseProperty);
                    this.Complete(false);
                }
                catch (WebException we)
                {
                    var responseProperty = new HttpResponseMessageProperty();
                    responseProperty.Headers.Add(we.Response.Headers);
                    responseProperty.StatusCode = ((HttpWebResponse) we.Response).StatusCode;
                    responseProperty.StatusDescription = ((HttpWebResponse) we.Response).StatusDescription;
                    responseProperty.SuppressEntityBody = true;
                    this.SignResponse(responseProperty);
                    this.result = Message.CreateMessage(this.request.Version, String.Empty);
                    this.result.Properties.Add(HttpResponseMessageProperty.Name, responseProperty);
                    this.Complete(false);
                }
                catch (Exception e)
                {
                    this.Complete(false, e);
                }
            }

            void MapRequestHeaders(HttpRequestMessageProperty rm)
            {
                foreach (var hdr in rm.Headers.AllKeys)
                {
                    switch (hdr.ToUpperInvariant())
                    {
                        case "AUTHORIZATION":
                        case "CONNECTION":
                        case "KEEP-ALIVE":
                        case "DATE":
                        case "TRANSFER-ENCODING":
                        case "VIA":
                        case "HOST":
                        case "EXPECT":
                            break;
                        case "CONTENT-LENGTH":
                            long cl;
                            if (!string.IsNullOrWhiteSpace(rm.Headers[hdr]) &&
                                long.TryParse(rm.Headers[hdr], out cl))
                            {
                                this.forwarderRequest.ContentLength = cl;
                            }
                            break;
                        case "ACCEPT":
                            this.forwarderRequest.Accept = rm.Headers[hdr];
                            break;
                        case "CONTENT-TYPE":
                            this.forwarderRequest.ContentType = rm.Headers[hdr];
                            break;
                        case "IF-MODIFIED-SINCE":
                            this.forwarderRequest.IfModifiedSince = DateTime.Parse(rm.Headers[hdr]);
                            break;
                        case "REFERER":
                            this.forwarderRequest.Referer = rm.Headers[hdr];
                            break;
                        case "USER-AGENT":
                            this.forwarderRequest.UserAgent = rm.Headers[hdr];
                            break;
                        default:
                            this.forwarderRequest.Headers.Add(hdr, rm.Headers[hdr]);
                            break;
                    }
                }
            }

            public static Message End(IAsyncResult ar)
            {
                End<HandleRequestAsyncResult>(ar);
                var hrar = (HandleRequestAsyncResult) ar;
                return hrar.result;
            }

            void CompleteWithError(HttpStatusCode statusCode, bool completedSynchronously)
            {
                this.result = Message.CreateMessage(this.request.Version, String.Empty);
                var rsp = new HttpResponseMessageProperty {StatusCode = statusCode, SuppressEntityBody = true};
                this.result.Properties.Add(HttpResponseMessageProperty.Name, rsp);
                this.Complete(completedSynchronously);
            }

            void SignResponse(HttpResponseMessageProperty wq)
            {
                var memStream = new MemoryStream();
                var signedHeaders = String.Empty;
                wq.Headers.Add("Generated", DateTime.UtcNow.ToString("R"));

                foreach (var headerName in wq.Headers.AllKeys)
                {
                    switch (headerName.ToUpperInvariant())
                    {
                        case "WWW-AUTHENTICATE":
                        case "TRANSFER-ENCODING":
                        case "CONTENT-LENGTH":
                        case "CONTENT-TYPE":
                        case "SERVER":
                        case "DATE":
                            break;
                        default:
                            signedHeaders += headerName + "&";
                            var headerValue = wq.Headers[headerName];
                            var buffer = Encoding.UTF8.GetBytes(headerValue);
                            memStream.Write(buffer, 0, buffer.Length);
                            break;
                    }
                }
                memStream.Flush();

                var hmaci = new HMACSHA256(signingKey);
                var hmac = hmaci.ComputeHash(memStream.ToArray());
                signedHeaders += Uri.EscapeDataString(Convert.ToBase64String(hmac));
                wq.Headers.Add("Signature", signedHeaders);
            }

            bool ValidateRequest(HttpRequestMessageProperty wr)
            {
                var signatureHeader = wr.Headers["Authorization"];
                if (signatureHeader == null ||
                    signatureHeader.Length <= 0)
                {
                    return false;
                }
                var signature = signatureHeader.Split(' ');
                if (signature.Length != 2 ||
                    !signature[0].Equals("SignedHeaders"))
                {
                    return false;
                }

                var memStream = new MemoryStream();
                var signatureElements = signature[1].Split('&');
                for (var i = 0; i < signatureElements.Length - 1; i++)
                {
                    var headerValue = wr.Headers[signatureElements[i]];
                    var buffer = Encoding.UTF8.GetBytes(headerValue);
                    memStream.Write(buffer, 0, buffer.Length);
                }
                memStream.Flush();

                var hmaci = new HMACSHA256(signingKey);
                var hmac = hmaci.ComputeHash(memStream.ToArray());
                var signatureElement = Uri.UnescapeDataString(signatureElements[signatureElements.Length - 1]);
                var fmac = Convert.FromBase64String(signatureElement);
                if (hmac.Length !=
                    fmac.Length)
                {
                    return false;
                }
                for (var i = 0; i < hmac.Length; i++)
                {
                    if (fmac[i] !=
                        hmac[i])
                    {
                        return false;
                    }
                }
                return true;
            }
        }
    }

    class DefaultOperationAttribute : Attribute, IContractBehavior, IDispatchOperationSelector
    {
        readonly string defaultOperation;

        public DefaultOperationAttribute(string defaultOperation)
        {
            this.defaultOperation = defaultOperation;
        }

        public void Validate(ContractDescription contractDescription, ServiceEndpoint endpoint)
        {
        }

        public void ApplyDispatchBehavior(ContractDescription contractDescription, ServiceEndpoint endpoint, DispatchRuntime dispatchRuntime)
        {
            dispatchRuntime.OperationSelector = this;
        }

        public void ApplyClientBehavior(ContractDescription contractDescription, ServiceEndpoint endpoint, ClientRuntime clientRuntime)
        {
        }

        public void AddBindingParameters(ContractDescription contractDescription, ServiceEndpoint endpoint, BindingParameterCollection bindingParameters)
        {
        }

        public string SelectOperation(ref Message message)
        {
            return this.defaultOperation;
        }
    }
}