﻿// 
//  (c) Microsoft Corporation. See LICENSE.TXT file for licensing details
//  
namespace BackendWebRole
{
    using System;
    using System.Configuration;
    using System.Linq;
    using System.Net;
    using System.ServiceModel;
    using System.ServiceModel.Web;
    using Microsoft.ServiceBus;
    using Microsoft.ServiceBus.AccessControlExtensions;
    using Microsoft.ServiceBus.Messaging;

    [ServiceContract(Namespace = "")]
    public class ProvisioningService
    {
        const string DevicesTopicPath = "devices";
        const string EventsTopicPath = "events";
        static readonly AccessControlSettings AccessControlSettings;
        static readonly string ManagementKey;
        static readonly string NamespaceName;
        static readonly byte[] signingKey;
        static readonly bool usePlainHttpMode;
        static Random rnd = new Random();


        static ProvisioningService()
        {
            bool.TryParse(ConfigurationManager.AppSettings["usePlainHttpMode"], out usePlainHttpMode);
            signingKey = Convert.FromBase64String(ConfigurationManager.AppSettings["sharedSignature"]);
            NamespaceName = ConfigurationManager.AppSettings["serviceBusNamespace"];
            ManagementKey = ConfigurationManager.AppSettings["managementKey"];
            AccessControlSettings = new AccessControlSettings(NamespaceName, ManagementKey);
        }

        [OperationContract, WebInvoke(Method = "POST", UriTemplate = "/setup")]
        public void SetupDevice()
        {
            var rcx = WebOperationContext.Current.OutgoingResponse;
            var qcx = WebOperationContext.Current.IncomingRequest;
            var id = qcx.Headers["P-DeviceId"];
            if (this.CheckAllowList(id))
            {
                try
                {
                    var deviceConfig = new DeviceConfig();
                    CreateServiceIdentity(ref deviceConfig);
                    CreateAndSecureEntities(ref deviceConfig);

                    rcx.Headers["P-DeviceAccount"] = deviceConfig.DeviceAccount;
                    rcx.Headers["P-DeviceKey"] = deviceConfig.DeviceKey;
                    rcx.Headers["P-DeviceSubscriptionUri"] = deviceConfig.DeviceSubscriptionUri;
                    rcx.Headers["P-EventSubmissionUri"] = deviceConfig.EventSubmissionUri;
                    rcx.StatusCode = HttpStatusCode.OK;
                    rcx.SuppressEntityBody = true;
                }
                catch (Exception)
                {
                    rcx.StatusCode = HttpStatusCode.InternalServerError;
                    rcx.SuppressEntityBody = true;
                }
            }
            else
            {
                rcx.StatusCode = HttpStatusCode.Forbidden;
                rcx.SuppressEntityBody = true;
            }
        }

        static void CreateAndSecureEntities(ref DeviceConfig deviceConfig)
        {
            var namespaceUri = ServiceBusEnvironment.CreateServiceUri(Uri.UriSchemeHttps, NamespaceName, string.Empty);
            var nsMgr = new NamespaceManager(namespaceUri, TokenProvider.CreateSharedSecretTokenProvider("owner", ManagementKey));
            var ruleDescription = new SqlFilter(string.Format("DeviceId='{0}' OR Broadcast=true", deviceConfig.DeviceAccount));
            var subscription = nsMgr.CreateSubscription(DevicesTopicPath, deviceConfig.DeviceAccount, ruleDescription);

            if (usePlainHttpMode)
            {
                deviceConfig.EventSubmissionUri = "/Passthrough.svc/" + EventsTopicPath;
                deviceConfig.DeviceSubscriptionUri = "/Passthrough.svc/" + SubscriptionClient.FormatSubscriptionPath(subscription.TopicPath, subscription.Name);
            }
            else
            {
                deviceConfig.EventSubmissionUri = new Uri(namespaceUri, EventsTopicPath).AbsoluteUri;
                deviceConfig.DeviceSubscriptionUri =
                    new Uri(namespaceUri, SubscriptionClient.FormatSubscriptionPath(subscription.TopicPath, subscription.Name)).AbsoluteUri;
                GrantSendOnEventTopic(deviceConfig);
                GrantListenOnDeviceSubscription(deviceConfig);
            }
        }

        static void GrantSendOnEventTopic(DeviceConfig deviceConfig)
        {
            var settings = new AccessControlSettings(NamespaceName, ManagementKey);
            var topicUri = ServiceBusEnvironment.CreateServiceUri(Uri.UriSchemeHttp, NamespaceName, EventsTopicPath);
            var list = NamespaceAccessControl.GetAccessControlList(topicUri, settings);
            var identityReference = IdentityReference.CreateServiceIdentityReference(deviceConfig.DeviceAccount);
            var existing = list.FirstOrDefault((r) => r.Condition.Equals(identityReference) && r.Right.Equals(ServiceBusRight.Send));
            if (existing == null)
            {
                list.AddRule(identityReference, ServiceBusRight.Send);
                list.SaveChanges();
            }
        }

        static void GrantListenOnDeviceSubscription(DeviceConfig deviceConfig)
        {
            var settings = new AccessControlSettings(NamespaceName, ManagementKey);
            var subscriptionUri = ServiceBusEnvironment.CreateServiceUri(
                                                                         Uri.UriSchemeHttp,
                                                                         NamespaceName,
                                                                         SubscriptionClient.FormatSubscriptionPath(DevicesTopicPath, deviceConfig.DeviceAccount));
            var list = NamespaceAccessControl.GetAccessControlList(subscriptionUri, settings);
            var identityReference = IdentityReference.CreateServiceIdentityReference(deviceConfig.DeviceAccount);
            var existing = list.FirstOrDefault((r) => r.Condition.Equals(identityReference) && r.Right.Equals(ServiceBusRight.Listen));
            if (existing == null)
            {
                list.AddRule(identityReference, ServiceBusRight.Listen);
                list.SaveChanges();
            }
        }

        static void CreateServiceIdentity(ref DeviceConfig deviceConfig)
        {
            var name = Guid.NewGuid().ToString("N");
            if (!usePlainHttpMode)
            {
                var identity = AccessControlServiceIdentity.Create(AccessControlSettings, name);
                identity.Key = signingKey;
                identity.Save();

                deviceConfig.DeviceAccount = identity.Name;
                deviceConfig.DeviceKey = identity.GetKeyAsBase64();
            }
            else
            {
                deviceConfig.DeviceAccount = name;
                deviceConfig.DeviceKey = Convert.ToBase64String(signingKey);
            }
        }


        bool CheckAllowList(string id)
        {
            return true;
        }
    }
}