// 
//  (c) Microsoft Corporation. See LICENSE.TXT file for licensing details
//  
namespace Microsoft.ServiceBus.Micro
{
    using System;
    using System.Collections;
    using System.IO;
    using System.Net;
    using System.Text;
    using ElzeKool;

    public class TokenProvider
    {
        readonly string issuerName;
        readonly string issuerSecret;
        readonly byte[] issuerSecretBytes;
        Hashtable cachedTokens;

        public TokenProvider(string issuerName, string issuerSecret)
        {
            this.issuerName = issuerName;
            this.issuerSecret = issuerSecret;
            this.issuerSecretBytes = ConvertBase64.FromBase64String(issuerSecret);
            this.cachedTokens = new Hashtable();
        }

        public virtual string GetToken(string serviceNamespace, string acsHostName, string sbHostName, string path)
        {
            var acsEndpoint = "https://" + serviceNamespace + "-sb." + acsHostName + "/WRAPv0.9/";
            var realm = "http://" + serviceNamespace + "." + sbHostName + path;
            var values = new Hashtable {{"wrap_name", this.issuerName}, {"wrap_password", this.issuerSecret}, {"wrap_scope", realm}};
            var responseString = ExecuteRequest(acsEndpoint, values);
            var responseProperties = responseString.Split('&');
            var tokenProperty = responseProperties[0].Split('=');
            var token = HttpUtility.UrlDecode(tokenProperty[1]);
            return "WRAP access_token=\"" + token + "\"";
        }

        static string ExecuteRequest(string webEndpoint, Hashtable values)
        {
            var wr = (HttpWebRequest) WebRequest.Create(webEndpoint);
            wr.Method = "POST";
            wr.ContentType = "application/x-www-form-urlencoded";
            var payload = string.Empty;
            foreach (string key in values.Keys)
            {
                payload += HttpUtility.UrlEncode(key) + "=" + HttpUtility.UrlEncode((string) values[key]) + "&";
            }
            var bytes = Encoding.UTF8.GetBytes(payload.Substring(0, payload.Length - 1));
            wr.ContentLength = bytes.Length;
            var rqs = wr.GetRequestStream();
            rqs.Write(bytes, 0, bytes.Length);
            rqs.Close();

            var wq = wr.GetResponse();
            var wrs = wq.GetResponseStream();


            var responseString = new StreamReader(wrs).ReadToEnd();
            wq.Close();
            return responseString;
        }

        public string GetToken(Uri serviceNamespace)
        {
            var host = serviceNamespace.Host;
            var @namespace = host.Substring(0, host.IndexOf("."));
            return this.GetToken(@namespace, "accesscontrol.windows.net", "servicebus.windows.net", serviceNamespace.AbsolutePath);
        }

        public void SignRequest(HttpWebRequest wq)
        {
            var memStream = new MemoryStream();
            var signedHeaders = String.Empty;
            foreach (var headerName in wq.Headers.AllKeys)
            {
                signedHeaders += headerName + "&";
                var headerValue = wq.Headers[headerName];
                var buffer = Encoding.UTF8.GetBytes(headerValue);
                memStream.Write(buffer, 0, buffer.Length);
            }
            memStream.Flush();

            var hmac = SHA.computeHMAC_SHA256(this.issuerSecretBytes, memStream.ToArray());
            var base64String = ConvertBase64.ToBase64String(hmac);
            var signature = string.Empty;
            for (var i = 0; i < base64String.Length; i++)
            {
                if (base64String[i] == '!')
                {
                    signature += '+';
                }
                else if (base64String[i] == '*')
                {
                    signature += '/';
                }
                else
                {
                    signature += base64String[i];
                }
            }
            signedHeaders += HttpUtility.UrlEncode(signature);
            wq.Headers.Add("Authorization", "SignedHeaders " + signedHeaders);
        }

        public bool ValidateResponse(HttpWebResponse wr)
        {
            var signature = wr.Headers["Signature"];
            if (signature == null ||
                signature.Length <= 0)
            {
                return false;
            }

            var memStream = new MemoryStream();
            var signatureElements = signature.Split('&');
            for (var i = 0; i < signatureElements.Length - 1; i++)
            {
                var headerValue = wr.Headers[signatureElements[i]];
                var buffer = Encoding.UTF8.GetBytes(headerValue);
                memStream.Write(buffer, 0, buffer.Length);
            }
            memStream.Flush();

            var hmac = SHA.computeHMAC_SHA256(this.issuerSecretBytes, memStream.ToArray());

            var base64String = HttpUtility.UrlDecode(signatureElements[signatureElements.Length - 1]);
            var signatureString = string.Empty;
            for (var i = 0; i < base64String.Length; i++)
            {
                if (base64String[i] == '/')
                {
                    signatureString += '*';
                }
                else if (base64String[i] == '+')
                {
                    signatureString += '!';
                }
                else
                {
                    signatureString += base64String[i];
                }
            }
            var fmac = ConvertBase64.FromBase64String(signatureString);
            if (hmac.Length !=
                fmac.Length)
            {
                return false;
            }
            for (var i = 0; i < hmac.Length; i++)
            {
                if (fmac[i] !=
                    hmac[i])
                {
                    return false;
                }
            }
            return true;
        }
    }
}