﻿// 
//  (c) Microsoft Corporation. See LICENSE.TXT file for licensing details
//  
using GT = Gadgeteer;
using GTM = Gadgeteer.Modules;

namespace ServiceBusApp
{
    using System;
    using System.IO;
    using System.Net;
    using Gadgeteer.Modules.GHIElectronics;
    using Gadgeteer.Modules.Seeed;
    using GT;
    using GTM;
    using Microsoft.ServiceBus.Micro;
    using Microsoft.SPOT;
    using Microsoft.SPOT.Hardware;
    using Microsoft.SPOT.Net.Security;

    public partial class Program
    {
        readonly string deviceId = String.Empty;
        readonly ApplicationSettings settings;
        X509Certificate[] caCerts;
        double lastHumidityReading;
        double lastTemperatureReading;
        MessagingClient messagingClient;
        string serverAddress = "http://cvdevices.cloudapp.net";
        Timer temperatureCheckTimer;
        TokenProvider tokenProvider;

        public Program()
        {
            this.settings = GetSettings();
        }

        void ProgramStarted()
        {
            this.InitializeNetwork();
            this.InitializeTemperatureSensor();
            this.InitializeJoystick();
        }

        void InitializeTemperatureSensor()
        {
            this.temperatureCheckTimer = new Timer(5000);
            this.temperatureCheckTimer.Tick += (t) => this.temperatureHumidity.RequestMeasurement();
            this.temperatureCheckTimer.Start();
            this.temperatureHumidity.MeasurementComplete += this.TemperatureHumidityMeasurementComplete;
        }

        void InitializeJoystick()
        {
            this.joystick.JoystickPressed += this.JoystickPressed;
        }


        void JoystickPressed(Joystick sender, Joystick.JoystickState state)
        {
            this.temperatureCheckTimer.Stop();
            var jStick = this.joystick.GetJoystickPostion();
            if (jStick.Y < .3 ||
                jStick.X < .3)
            {
                this.settings.TargetTemperature -= .5;
                StoreSettings(this.settings);
            }
            else if (jStick.Y > .7 ||
                     jStick.X > .7)
            {
                this.settings.TargetTemperature += .5;
                StoreSettings(this.settings);
            }
            this.RedrawDisplay();
            this.temperatureHumidity.RequestMeasurement();
            this.temperatureCheckTimer.Start();
        }

        void TemperatureHumidityMeasurementComplete(TemperatureHumidity sender, double temperature, double relativeHumidity)
        {
            var targetTemp = this.settings.TargetTemperature;
            this.lastTemperatureReading = temperature;
            this.lastHumidityReading = relativeHumidity;
            this.relays.Relay1 = (this.lastTemperatureReading > targetTemp);
            this.RedrawDisplay();

            if (this.settings.NetworkProvisioningCompleted &&
                DateTime.UtcNow - this.settings.LastServerUpdate > TimeSpan.FromTicks(TimeSpan.TicksPerMinute))
            {
                this.settings.LastServerUpdate = DateTime.UtcNow;
                this.SendEvent(this.lastTemperatureReading, this.lastHumidityReading);
                this.ProcessCommands();
            }
        }

        void ProcessCommands()
        {
            SimpleMessage cmd = null;
            try
            {
                do
                {
                    cmd = this.messagingClient.Receive(TimeSpan.Zero, ReceiveMode.ReceiveAndDelete);
                    if (cmd != null &&
                        cmd.Properties.Contains("Command"))
                    {
                        var commandType = (string) cmd.Properties["Command"];
                        switch (commandType)
                        {
                            case "SetTemperature":
                                if (cmd.Properties.Contains("Parameter"))
                                {
                                    this.settings.TargetTemperature = double.Parse((string) cmd.Properties["Parameter"]);
                                    this.RedrawDisplay();
                                    this.temperatureHumidity.RequestMeasurement();
                                    StoreSettings(this.settings);
                                }
                                break;
                        }
                    }
                }
                while (cmd != null);
            }
            catch (Exception e)
            {
                Debug.Print(e.ToString());
            }
        }

        void SendEvent(double d, double lastHumidityReading1)
        {
            try
            {
                this.messagingClient.Send(
                                          new SimpleMessage()
                                              {Properties = {{"Temperature", d}, {"Humidity", lastHumidityReading1}, {"DeviceId", this.settings.DeviceAccount}}});
            }
            catch (Exception e)
            {
                Debug.Print(this.ethernet.ToString());
            }
        }

        void InitializeNetwork()
        {
            this.ethernet.UseDHCP();
            this.ethernet.NetworkUp += this.NetworkAvailable;
        }

        bool PerformProvisioning(string serverAddress)
        {
            this.oledDisplay.SimpleGraphics.ClearNoRedraw();
            this.oledDisplay.SimpleGraphics.DisplayText("Service Bus A/C", Resources.GetFont(Resources.FontResources.NinaB), Color.White, 10, 10);
            this.oledDisplay.SimpleGraphics.DisplayText("Provisioning", Resources.GetFont(Resources.FontResources.NinaB), Color.Red, 10, 30);

            try
            {
                var wr = WebRequest.Create(serverAddress + "/Provisioning.svc/setup");
                wr.Method = "POST";
                wr.ContentLength = 0;
                wr.Headers.Add("P-DeviceId", this.deviceId);

                using (var wq = (HttpWebResponse) wr.GetResponse())
                {
                    if (wq.StatusCode ==
                        HttpStatusCode.OK)
                    {
                        this.settings.DeviceAccount = wq.Headers["P-DeviceAccount"];
                        this.settings.DeviceKey = wq.Headers["P-DeviceKey"];

                        var subUriString = wq.Headers["P-DeviceSubscriptionUri"];
                        if (subUriString != null &&
                            subUriString.Length > 0)
                        {
                            this.settings.DeviceSubscriptionUri = subUriString[0] == '/' ? new Uri(serverAddress + subUriString) : new Uri(subUriString);
                        }
                        var eventUriString = wq.Headers["P-EventSubmissionUri"];
                        if (eventUriString != null &&
                            eventUriString.Length > 0)
                        {
                            this.settings.EventSubmissionUri = eventUriString[0] == '/' ? new Uri(serverAddress + eventUriString) : new Uri(eventUriString);
                        }
                        this.settings.NetworkProvisioningCompleted = true;
                        return true;
                    }
                    else
                    {
                        Debug.Print(new StreamReader(wq.GetResponseStream()).ReadToEnd());
                    }
                }
            }
            catch (Exception e)
            {
                return false;
            }
            return false;
        }

        static ApplicationSettings GetSettings()
        {
            var data = ExtendedWeakReference.RecoverOrCreate(
                                                             typeof (ApplicationSettings),
                                                             0,
                                                             ExtendedWeakReference.c_SurviveBoot | ExtendedWeakReference.c_SurvivePowerdown);
            var settings = data.Target as ApplicationSettings;
            if (settings == null)
            {
                data.Target = settings = ApplicationSettings.Defaults;
            }
            return settings;
        }

        static void StoreSettings(ApplicationSettings settings)
        {
            var data = ExtendedWeakReference.RecoverOrCreate(
                                                             typeof (ApplicationSettings),
                                                             0,
                                                             ExtendedWeakReference.c_SurviveBoot | ExtendedWeakReference.c_SurvivePowerdown);
            data.Target = settings;
            data.PushBackIntoRecoverList();
        }


        void NetworkAvailable(Module.NetworkModule sender, Module.NetworkModule.NetworkState state)
        {
            ConvertBase64.ToBase64String(this.ethernet.NetworkSettings.PhysicalAddress);
            if (state == Module.NetworkModule.NetworkState.Up)
            {
                try
                {
                    Utility.SetLocalTime(NtpClient.GetNetworkTime());
                }
                catch
                {
                    // swallow any timer exceptions
                }

                if (!this.settings.NetworkProvisioningCompleted)
                {
                    if (!this.PerformProvisioning(this.serverAddress))
                    {
                        return;
                    }
                }
                if (this.settings.NetworkProvisioningCompleted)
                {
                    this.tokenProvider = new TokenProvider(this.settings.DeviceAccount, this.settings.DeviceKey);
                    this.messagingClient = new MessagingClient(this.settings.EventSubmissionUri, this.tokenProvider);
                }
            }
        }

        void RedrawDisplay()
        {
            var targetTemp = this.settings.TargetTemperature;

            this.oledDisplay.SimpleGraphics.ClearNoRedraw();
            this.oledDisplay.SimpleGraphics.DisplayText(DateTime.UtcNow.ToString(), Resources.GetFont(Resources.FontResources.NinaB), Color.Red, 10, 10);
            this.oledDisplay.SimpleGraphics.DisplayText(
                                                        "Target: " + targetTemp.ToString("F2"),
                                                        Resources.GetFont(Resources.FontResources.NinaB),
                                                        Color.Red,
                                                        20,
                                                        30);
            this.oledDisplay.SimpleGraphics.DisplayText(
                                                        "Temp: " + this.lastTemperatureReading.ToString("F2"),
                                                        Resources.GetFont(Resources.FontResources.NinaB),
                                                        Color.Red,
                                                        20,
                                                        50);
            this.oledDisplay.SimpleGraphics.DisplayText(
                                                        "Humidity: " + this.lastHumidityReading.ToString("F2"),
                                                        Resources.GetFont(Resources.FontResources.NinaB),
                                                        Color.Red,
                                                        20,
                                                        70);
            if (this.relays.Relay1)
            {
                this.oledDisplay.SimpleGraphics.DisplayText("Fan On", Resources.GetFont(Resources.FontResources.NinaB), Color.Red, 20, 100);
            }
            else
            {
                this.oledDisplay.SimpleGraphics.DisplayText("Fan Off", Resources.GetFont(Resources.FontResources.NinaB), Color.Blue, 20, 100);
            }
        }
    }
}